/*

  Program Burgers.ccp:
  Program to integrate the Burger's partial differential equation
  through a first order forward Euler scheme with a spectral method (FFT).

  d u / dt = - u * d u / dx - v * d^2 u / dx^2

  in an interval [a,b], with periodic boundary conditions,
  in the following cases:
  1) constant c;
  2) Burgers' equation.


  The program reads the input data from file input.data.

 */

#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <fftw3.h>
using namespace std;


// Output function, defined below the main program

void output( int, double, double *, double *, int );


/*

  Main program

*/

int main()
{

  /*
    Read initial data from file
  */

  ifstream finp;
  finp.open( "input.data" );

  // Read number of grid-points

  long int numPoints;
  finp >> numPoints;

  // Read the time step

  double dt;
  finp >> dt;

  // Read total time of the simulation

  double tEnd;
  finp >> tEnd;

  // Read time interval between the outputs

  double tPrint;
  finp >> tPrint;

  // Velocity parameter c (if not Burgers!)
  
  double c;
  finp >> c;

  // Dissipation parameter nu
  
  double nu;
  finp >> nu;

  // Generic parameter k (wavenumber of the initial condition)

  long int k;
  finp >> k;

  // Type of equation to integrate:
  // teq = 1 => constant velocity c = 1;
  // teq = 2 => Burgers' equation: c(x) = u(x);

  int teq;
  finp >> teq;

  finp.close();


  // Definition of the variables for the spatial grid a=0, b=2*pi

  double pi = acos( -1.0 );
  double a, b;
  a = 0.0;
  b = 2.0 * pi;


  // Variables for time loop

  long nEnd   = long( tEnd / dt );
  long nPrint = long( tPrint / dt );


  // Vectors for fftw and quantities

  double *in;
  fftw_complex *out;
  fftw_plan pf, pb;
  double *x, *u;
  fftw_complex *uu, *d1f, *d2f;

  // Builds up the vectors for the quantities

  int n = numPoints - 1;
  int n2 = n / 2 + 1;     // Nyquist number + 1
  in = ( double * ) fftw_malloc( sizeof( double ) * n );
  out = ( fftw_complex * ) fftw_malloc( sizeof( fftw_complex ) * n2 );

  x = ( double * ) fftw_malloc( sizeof( double ) * n );
  u = ( double * ) fftw_malloc( sizeof( double ) * n );
  uu  = ( fftw_complex * ) fftw_malloc( sizeof( fftw_complex ) * n2 );
  d1f = ( fftw_complex * ) fftw_malloc( sizeof( fftw_complex ) * n2 );
  d2f = ( fftw_complex * ) fftw_malloc( sizeof( fftw_complex ) * n2 );


  // Create the forward and backward transform planes
  // Notice the swap of in and out in the backward plan!

  unsigned int flags = FFTW_ESTIMATE || FFTW_DESTROY_INPUT;
  pf = fftw_plan_dft_r2c_1d( n, in, out, flags );
  pb = fftw_plan_dft_c2r_1d( n, out, in, flags );


  // Define the grid-points and the initial condition

  double tn = 0.0;
  int numPrint = 0;
  double dx = ( b - a ) / n;
  for ( int j = 0; j < n; j++ )
    {
      x[j] = a + j * dx;
      u[j] = sin( k * x[j] );
    }


  // Output of the initial condition

  output( numPrint, tn, x, u, n );

  
  // Time loop

  for ( long int nt = 1; nt <= nEnd; nt++ )
    {

      // First step for the Runge-Kutta time scheme

      // Compute the first and second derivative

      // Computes the forward transform
  
      for ( int i = 0; i < n; i++ ) in[i] = u[i];
      fftw_execute( pf );

      // Normalize the transform!

      for ( int k = 0; k < n2; k++ )
	{
	  uu[k][0] = out[k][0] / n;     // Normalize the real parts of the output vector
	  uu[k][1] = out[k][1] / n;     // Normalize the imaginary parts of the output vector
	}
      
      for ( int k = 0; k < n2 - 1; k++ )
	{
	  d1f[k][0] = - k * uu[k][1];
	  d1f[k][1] = k * uu[k][0];
	  d2f[k][0] = - k * k * uu[k][0];
	  d2f[k][1] = - k * k * uu[k][1];
	}

      // Puts to zero the last harmonics (n2)!

      d1f[n2-1][0] = 0.0;
      d1f[n2-1][1] = 0.0;
      d2f[n2-1][0] = 0.0;
      d2f[n2-1][1] = 0.0;
      
      // Compute the value of c * du/dx or u*du/dx in the Fourier space
      // according to the type of equation (teq = 1 or teq = 2)

      if ( teq == 1 )
	{
	  for ( int k = 0; k < n2; k++ )
	    {
	      out[k][0] = c * d1f[k][0];     // Computes c * du/dx, real part
	      out[k][1] = c * d1f[k][1];     // Computes c * du/dx, imaginary part
	    }
	}


      if ( teq == 2 )
	{
	  // Computes the backward transform of du

	  for ( int k = 0; k < n2; k++ )
	    {
	      out[k][0] = d1f[k][0];
	      out[k][1] = d1f[k][1];
	    }
	  fftw_execute( pb );
      
	  // Computes u*du/dx in physical space
	  
	  for ( int i = 0; i < n; i++ )
	    in[i] = u[i] * in[i];
	  
	  // Computes udu in spectral space

	  fftw_execute( pf );

	  // Normalize the transform!

	  for ( int k = 0; k < n2; k++ )
	    {
	      out[k][0] = out[k][0] / n;     // Normalize the real parts of the output vector
	      out[k][1] = out[k][1] / n;     // Normalize the imaginary parts of the output vector
	    }
	}
	  
      
      // Compute the solution at next time-step

      for ( int k = 0; k < n2; k++ )
	{
	  uu[k][0] = uu[k][0] + dt * ( - out[k][0] + nu * d2f[k][0] );
	  uu[k][1] = uu[k][1] + dt * ( - out[k][1] + nu * d2f[k][1] );
	}
	
      // Computes the solution in the physical space
      
      // Computes the backward transform of u

      for ( int k = 0; k < n2; k++ )
	{
	  out[k][0] = uu[k][0];
	  out[k][1] = uu[k][1];
	}
      fftw_execute( pb );
      
      for ( int i = 0; i < n; i++ )
	u[i] = in[i];
      

      // Prints the outputs each Tprint time intervals

      tn = nt * dt;
      if ( ( nt % nPrint ) == 0 )
	{
	  numPrint++;
	  output( numPrint, tn, x, u, n );
	}

    }

}




/*

  Function output

*/

void output( int np, double tt, double *x, double *f, int N )
{
  string fname = "out-" + to_string( np );
  string timestr = "# t = " + to_string( tt );
  ofstream fout;
  fout.open( fname );
  fout.precision( 15 );
  fout << timestr << endl;
  for ( int j = 0; j < N; j++ )
    fout << x[j] << "   " << f[j] << endl;
  fout.close();
}
